/* This file was generated by syncqt. */
#ifndef QT_QTWEBCHANNELQUICK_VERSION_H
#define QT_QTWEBCHANNELQUICK_VERSION_H

#define QTWEBCHANNELQUICK_VERSION_STR "6.11.0"

#define QTWEBCHANNELQUICK_VERSION 0x0601100

#endif // QT_QTWEBCHANNELQUICK_VERSION_H
