// Copyright (C) 2016 Klarälvdalens Datakonsult AB, a KDAB Group company, info@kdab.com, author Milian Wolff <milian.wolff@kdab.com>
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
// Qt-Security score:significant reason:default

#ifndef QTWEBCHANNEL_H
#define QTWEBCHANNEL_H

#include <QtCore/qglobal.h>
#include <QtWebChannel/qtwebchannelexports.h>

#endif // QTWEBCHANNEL_H
